function [ Kcl, Jcl ] = clopt( sysStruct, probStruct, K0, varargin )
%CLOPT Closed-loop H_2 optimization
%   [Kcl,Jcl] = clopt( sysStruct, probStruct, K0, 'method' )
%   
%   Outputs:
%   Kcl         - Closed-loop optimal controller.
%   Jcl         - Objective function evaluation for Kcl.
%
%   Inputs:
%   sysStruct   - System struct (continuous or discrete). 
%                 Generated using ss() or sysStruct().
%   probStruct  - Control problem struct holding Q,R,P,N.
%                 Generated using probStruct().
%   K0          - Initial guess on controller (Negative feedback).
%   method      - Cost function formulation (Optional). 
%                 Default setting is 'lft'.
% 
%   clopt( sysStruct, probStruct, K0, 'lft' ): 
%   Minimizes the norm of the lower fractional transform of the 
%   generalized system, i.e. norm(lft(P,-K)). Please note that norm
%   minimalization requires that eigenvalues are in the open left 
%   half-plane, or open unit disc in the discrete case.
%
%   clopt( sysStruct, probStruct, K0, 'impulse' ): 
%   minimizes the weighted impulse response of the system, 
%   i.e. norm(trace(M(K))), where M(K) is the objective function 
%   specified by probStruct. Type help calcM to get more information.

% Tests arguments
if (nargin < 3)
    display('Please supply enough arguments!');
    display('Usage: clopt( sysStruct, probStruct, K0, varargin )');
elseif (nargin == 3)
    method = 'lft';
elseif (nargin == 4)
    if (strcmp(varargin{1}, 'lft'))
        method = 'lft';
    elseif (strcmp(varargin{1}, 'impulse'))
        method = 'impulse';
    else
        display('Incorrect method supplied. Using lft by default!');
        method = 'lft';
    end
else
    display('Too many arguments supplied!');
    display('Usage: clopt( sysStruct, probStruct, K0, varargin )');
end

% Closed-loop H_2 optimization
if (strcmp(method,'impulse')) % Optimization using impulse response
    
    % J = x'Px + sum(0,N) (x'Qx + u'Ru) = x0'M(K)x0
    % Impulse response is given by: trace(M(K))
    display('Problem formulated using impulse response.');
    display('Solving closed-loop H_2 optimization...');

    f = inline('trace(calcM(sysStruct,probStruct,K))','sysStruct','probStruct','K');
    options = optimset('Display', 'iter', 'MaxFunEvals', 10000);
    [Kcl, Jcl] = fminunc(@(K) f(sysStruct,probStruct,K), K0, options);
    
else % Optimization using lft
    
    display('Problem formulated using lower fractional transform.');
    display('Solving closed-loop H_2 optimization...');
    
    if (isempty(probStruct.W) || isempty(probStruct.V))
        error('Please specify weight matrices W and/or V');
    end
    
    P = calcP(sysStruct, probStruct); % Calculates gen. plant
    
    if (clnorm(P, -K0) == Inf)
        error('Closed-loop norm is Inf. Optimization terminated!')
    end
    
    f = inline('clnorm(P,-K)', 'K', 'P'); % Note negative feedback, u = -Ky
    options = optimset('Display', 'iter', 'MaxFunEvals', 10000);
    [Kcl, Jcl] = fminunc(@(K) f(K,P), K0, options);
    
end


end

