% This example illustrates that even simple SISO systems may yield a 
% non-convex closed-loop optimization problem
close all
clear all
clc

warning off all

% Plant model
% A = [0.9 -0.1 0.2; 
%     0.1 0.7 -.4;
%     0 1 -0.2]; 
% B = [-0.5 0.3 0.1]'; 
% C = [1 1 1]; 
% D = 0;
% Ts = 0.5;
% system = ss(A,B,C,D,Ts);
% [nx,nu,ny] = sysdim(A,B,C);

% Lets go for this system :)
s = tf('s');
num = (0.1*s + 1)*(0.3*s + 1);
den = 10*s*(1*s + 1)*(1.5*s + 1);
G = num/den;
Gss = ss(G);
[A,B,C,D] = ssdata(Gss);

system = Gss;
[nx,nu,ny] = sysdim(A,B,C);

eig(A)
isControllable(A,B)
isObservable(A,C)

% Performance weights
%Q = eye(nx);
Q = 10*C'*C;
R = 0.01;
P = zeros(nx); %P = dlyap(A',Q); % For discrete systems
%P = lyap(A',Q);
W = eye(nx);
V = zeros(ny);
N = 200;
problem = probStruct(Q,R,P,N,W,V);
    
% generalized plant
P = calcP(system,problem);

% LQR controller. Used to find a suboptimal controller
Klqr = dlqr(A, B, Q, R); 

[Juu, Jud] = calcJuu(system, problem);
Kinit = inv(Juu)*Jud*pinv(C);
Kinit = Kinit(1);

Kinit2 = Klqr*pinv(C);

% Data vectors
%K = -10:0.1:10;
K = [-50:1:99 100:10:1000]; % works for the second system
% K = [0:10:1000 1000:100:1e5]; % works for the first system
m = length(K);

Jvec = zeros(m,1);
Mvec = zeros(m,1);

% disp('Finding objective functions...');
for i = 1:m
   Jvec(i) = clnorm(P,-K(i));
   %Mvec(i) = trace(calcM(system,problem,K(i)));
end

minima = find_minima(Jvec)

if minima > 1
    
    disp('Found a non-convex cost function! Saving data...');
    
    % save data
    timenow = datestr(now, 'dd-mm-yyyy_HH-MM-SS');
    filename = ['./data/SISO_Jcl_', timenow];
    %save(filename, 'Jvec', 'system', 'problem');
end

disp('Plotting objective functions...');
figure
plot(K, Jvec, 'Linewidth', 1.5);
xlim([K(1) K(end)]);
ylim([0 51]);
xlabel('Controller gain');
ylabel('Cost function');
l = legend('$J_{cl}$');
set(l, 'Interpreter', 'Latex')
set(l, 'FontSize', 12)

% figure
% plot(K,Mvec);
% xlim([K(1) K(end)]);
% ylim([0 51]);

% Bode plot
%bode(G)
[mag,phase,omega] = bode(G);
mag = 20*log10(mag(:));
phase = phase(:);

figure
subplot(211), semilogx(omega, mag, 'Linewidth', 1.5)
hold on, semilogx(omega, 0*ones(length(omega),1), '-k'), hold off
ylabel('Magnitude [dB]');
subplot(212), semilogx(omega, phase, 'Linewidth', 1.5)
hold on, semilogx(omega, -180*ones(length(omega),1), '-k'), hold off
ylim([-190 -88]);
set(gca,'Ytick',[-180:30:-90]);
ylabel('Phase [deg]');
xlabel('Frequency [rad/sec]');

