%% LQG control of a unstable SISO process with a RH-zero
% This example aims to highlight the difficulties of constructing an
% observer for non-minimum phase plants with small model errors
clc; clear all; close all

g1 = tf(-1,[1 -1]); % Unstable pole at s = 1
g2 = tf([1 -0.5],[1 0.5]); % RHP-zero at s = 2, and pole at s = -2
G = g1*g2; % Ideal plant without model errors

% State-space representation
Gss = ss(G);
[A,B,C,D] = ssdata(Gss);
%[nx,nu,ny] = sysdim(A,B,C);

% LQR controller
Q = diag([10 10]); % Q = diag(10,10)
R = 1; % R = 1
Kr = lqr(A,B,Q,R); % Negative, LQ optimal feedback u = -K*xhat
display('LQR poles: ')
eig(A-B*Kr) % Closed-loop eigenvalues with state feedback

% Observer design
W = 1; % Covariance of input disturbance, W = 1
V = .01; % Covariance of measurement noise, V = 0.01
Gest = ss(A, [B B], C, [D 0]); % Input disturbances and measurement noise
[kest,Kf] = kalman(Gest,W,V); % Computes Kalman gain
% p = [complex(-5,1) complex(-5,-1)]; % Desired observer poles
% Kf = place(A',C',p)';
display('Observer poles: ')
eig(A-Kf*C)

% Overall controller
Klqg = ss(A - Kf*C - B*Kr, Kf, -Kr, 0); % LQR + Observer (Skogestad p.347)
Klqg = tf(Klqg);
% display('Poles of LQG controller: ')
% pole(Klqg)

Gcl = feedback(G,Klqg,+1);
display('Poles of closed-loop system: ')
pole(Gcl) % Combination of LQR and observer poles

% Stability check
% Gloop = G*Klqg; % Loop transfer function (positive feedback)
% S = inv(1 - Gloop); S = minreal(S);
% T = inv(1 - Gloop)*Gloop; T = minreal(T);
% 
% pole(S)
% isstable(S)
% 
% pole(T)
% isstable(T)
% 
% pole(minreal(S*G))
% isstable(minreal(S*G))

% Plots nyquist diagram of the open-loop tf.
% The transfer function has two poles in the RHP, 
% by the nyquist stability criterion there must be 
% two counter clock-wise rotations around (-1, 0j)
% for the closed-loop system to be stable
% figure()
% nyquist(-G*Klqg);

% Its all good!

%% MPC controller
Ts = 0.1;
p = 10; % Prediction horizon
m = p; % Control horizon
weights = struct('Input',R,'InputRate',0,'Output',C*Q*C');
MV = struct('Min',-1000,'Max',1000);
mpc_obj = mpc(G,Ts,p,m,weights,MV);

%% Simulating the system with LQG control or MPC
wu = 0.01; % Input disturbance
wy = 0.01; % Measurement noise
tdelay = 0; % Time delay, treshold on 0.07
mpc_toggle = 0; % 0: LQG 1: MPC
simTime = 20;
sim('obsv_mdl.mdl', simTime); % LQG and noisy plant
%sim('obsv_mdl_with_MPC.mdl', simTime); % LQG and MPC
%sim('obsv_mdl2.mdl', simTime); % Only observer
%sim('obsv_mdl3.mdl', simTime); % LQR control, and passive observer

time = yout.time;
y = yout.signals.values(:,1);
yhat = yout.signals.values(:,2);

figure()
plot(time, y, 'Linewidth', 2);
hold on
plot(time, yhat, '-r', 'Linewidth', 2);
hold off
xlim([0 simTime]);
ylim([-10 10]);
ylabel('Estimation error');
xlabel('Time [s]');
l = legend('$y$', '$\hat{y}$');
set(l, 'Interpreter', 'Latex')
set(l, 'FontSize', 12)

figure()
plot(time, y - yhat, 'Linewidth', 2);
xlim([0 simTime]);
ylim([-10 10]);
ylabel('Estimation error');
xlabel('Time [s]');
l = legend('$e = y - \hat{y}$');
set(l, 'Interpreter', 'Latex')
set(l, 'FontSize', 12)

figure()
u = uout.signals.values;
plot(time, u, 'Linewidth', 2);
xlim([0 simTime]);
ylim([-100 100]);
ylabel('Input');
xlabel('Time [s]');
l = legend('$u$');
set(l, 'Interpreter', 'Latex')
set(l, 'FontSize', 12)

%% Only for observer without LQR
x1 = xout.signals.values(:,1);
x2 = xout.signals.values(:,2);
xhat1 = xhatout.signals.values(:,1);
xhat2 = xhatout.signals.values(:,2);

figure()
plot(time, x1 - xhat1, 'Linewidth', 3);
hold on
plot(time, x2 - xhat2, '-.k', 'Linewidth', 1.5);
hold off
xlim([0 simTime]);
ylim([-10 10]);
ylabel('State estimates');
xlabel('Time [s]');
l = legend('$e_1 = x_1 - \hat{x}_1$','$e_2 = x_2 - \hat{x}_2$');
set(l, 'Interpreter', 'Latex')
set(l, 'FontSize', 12)

