% This example presents the closed-loop norms of a P, PI, PID, and LQR
% controller on a second-order system. It is expected that the LQR and PID
% controller has the same norm, while the PI controller has a higher norm,
% followed by the P controller.

% Note: adding integrators makes control design difficult as the
% closed-loop norm can become infinite

clc
clear all
close all

% The general second-order system is described as 
% g(s) = k/(tau^2 s^2 + 2 tau zeta s + 1)
% where tau = 1 and zeta = [0,1]
% zeta < 1 gives an underdamped system
k = 1;
tau = 1;
zeta = 0.8;
g = tf(k, [tau^2, 2*tau*zeta, 1]);
G = ss(g); eig(G)
A = G.A; B = G.B;
C = G.C; D = G.D;
[nx,nu,ny] = sysdim(A,B,C);

% Discretization
% Note that d2c with zoh does not handle poles at zero z = 0, or negative
% real poles
Ts = 0.1;
Gdisc = c2d(G,Ts); eig(Gdisc)
Ad = Gdisc.A; Bd = Gdisc.B;
Cd = Gdisc.C; Dd = Gdisc.D;

% Problem formulation
Q = Cd'*Cd; %Q = diag([1 1]);
%Q = eye(2);
R = 1;
N = 100; % Horizon length
P = Q;
%P = dlyap(Ad',Q); % Solves the Lyapunov equation P = A'PA + Q

%% LQR controller (all states available)
Qlqr = Q;
Rlqr = R;
Plqr = P;
Wlqr = eye(nx);
Vlqr = zeros(nx);
probLQR = probStruct(Qlqr,Rlqr,Plqr,N,Wlqr,Vlqr);

Glqr = ss(Ad, Bd, eye(nx), zeros(nx,nu), Ts);

Klqr = dlqr(Ad,Bd,Qlqr,Rlqr);

Plqr = calcP(Glqr, probLQR);
Jlqr = clnorm(Plqr, -Klqr);
% Jlqr = trace(calcM(Glqr,probLQR,Klqr));

%% PID controller with limited derivative action
epsilon = 0.1; % Filter coefficient, numerical problems when set too low!

% Apid = [A zeros(nx,ny+ny);
%         C zeros(ny,ny+ny);
%         (1/epsilon)*C*A zeros(ny) -(1/epsilon)*eye(ny)
%         ];
% Bpid = [B; D; (epsilon*C*B-D)/(epsilon^2)];
% Cpid = blkdiag(C, eye(ny+ny)); 
% Dpid = [D; zeros(ny,nu); D/epsilon];

%Gpid = ss(Apid,Bpid,Cpid,Dpid); eig(Gpid) 

Gpid = sysaug(G, 'PID', 0, 'pos', epsilon);
Gpid = c2d(Gpid, Ts); % Discretizes

Apid = Gpid.A; Bpid = Gpid.B;
Cpid = Gpid.C; Dpid = Gpid.D;

[nx_pid, nu_pid, ny_pid] = sysdim(Apid, Bpid, Cpid);

% Note: if the integral action is not weighted, the marginal poles are not
% moved and the closed-loop system norm becomes Inf!
Qpid = Cpid'*diag([1 1e-3 0])*Cpid;
Rpid = R;
Ppid = Qpid; %Ppid = dlyap(Apid',Qpid);
Wpid = blkdiag(eye(nx),zeros(ny+ny)); % only weight on process states
Vpid = zeros(ny_pid);

probPID = probStruct(Qpid,Rpid,Ppid,N,Wpid,Vpid);

% Optimal measurement combination for PID control
[Juu, Jud] = calcJuu(Gpid, probPID);

Gu = [Dpid zeros(ny_pid,nu*(N-1));
      eye(nu*N)
      ];
Gd = [Cpid;
      zeros(nu*N,nx_pid)
      ];
  
Wd = eye(nx_pid);
Wn = blkdiag(zeros(ny_pid),zeros(nu*N)); % No noise

[H, Lwc_pid, Jopt_pid] = optmeas(Gu, Gd, Juu, Jud, Wd, Wn);

H1 = H(:,1:ny_pid);
H2 = H(:,ny_pid+1:end);
K = inv(H2)*H1;
K0_pid = K(1:nu,:); % Gives the relation u = K0*y, where y are the meas candidates

genPpid = calcP(Gpid,probPID);
J0_pid = clnorm(genPpid, -K0_pid);
%eig(Apid - Bpid*K0_pid*Cpid)
[Kpid, Jpid] = clopt(Gpid, probPID, K0_pid, 'lft');

%% PD controller with limited derivative action
epsilon = 0.1; % Filter coefficient, numerical problems when set too low!

Gpd = sysaug(G, 'PD', 0, 'pos', epsilon);
Gpd = c2d(Gpd, Ts); % Discretizes

Apd = Gpd.A; Bpd = Gpd.B;
Cpd = Gpd.C; Dpd = Gpd.D;

[nx_pd, nu_pd, ny_pd] = sysdim(Apd, Bpd, Cpd);

% Note: if the integral action is not weighted, the marginal poles are not
% moved and the closed-loop system norm becomes Inf!
Qpd = Cpd'*diag([1 0])*Cpd;
Rpd = R;
Ppd = Qpd; %Ppid = dlyap(Apid',Qpid);
Wpd = blkdiag(eye(nx),zeros(ny)); % only weight on process states
Vpd = zeros(ny_pd);

probPD = probStruct(Qpd,Rpd,Ppd,N,Wpd,Vpd);

% Optimal measurement combination for PID control
[Juu, Jud] = calcJuu(Gpd, probPD);

Gu = [Dpd zeros(ny_pd,nu*(N-1));
      eye(nu*N)
      ];
Gd = [Cpd;
      zeros(nu*N,nx_pd)
      ];
  
Wd = eye(nx_pd);
Wn = blkdiag(zeros(ny_pd),zeros(nu*N)); % No noise

[H, Lwc_pd, Jopt_pd] = optmeas(Gu, Gd, Juu, Jud, Wd, Wn);

H1 = H(:,1:ny_pd);
H2 = H(:,ny_pd+1:end);
K = inv(H2)*H1;
K0_pd = K(1:nu,:); % Gives the relation u = K0*y, where y are the meas candidates

genPpd = calcP(Gpd,probPD);
J0_pd = clnorm(genPpd, -K0_pd);
%eig(Apd - Bpd*K0_pd*Cpd)
[Kpd, Jpd] = clopt(Gpd, probPD, K0_pd, 'lft');

%% PI controller
% Api = [A zeros(nx,ny);
%        C zeros(ny,ny)
%        ];
% Bpi = [B; D];
% Cpi = blkdiag(C, eye(ny));
% Dpi = [D; zeros(ny,nu)];
% 
% Gpi = ss(Api,Bpi,Cpi,Dpi); eig(Gpi)
Gpi = sysaug(G, 'PI', 0, 'pos');
Gpi = c2d(Gpi, Ts); % Discretizes

Api = Gpi.A; Bpi = Gpi.B;
Cpi = Gpi.C; Dpi = Gpi.D;

[nx_pi, nu_pi, ny_pi] = sysdim(Api, Bpi, Cpi);

% Note: if the integral action is not weighted, the marginal poles are not
% moved and the closed-loop system norm becomes Inf!
Qpi = blkdiag(Q,1e-4);
Qpi = Cpi'*diag([1 1e-3])*Cpi;
Rpi = R;
Ppi = Qpi; %Ppi = dlyap(Api',Qpi);
Wpi = blkdiag(eye(nx), zeros(ny));
Vpi = zeros(ny_pi);
probPI = probStruct(Qpi,Rpi,Ppi,N,Wpi,Vpi);

% Optimal measurement combination for PI control
[Juu, Jud] = calcJuu(Gpi, probPI);

Gu = [Dpi zeros(ny_pi,nu*(N-1));
      eye(nu*N)
      ];
Gd = [Cpi;
      zeros(nu*N,nx_pi)
      ];

Wd = eye(nx_pi);
Wn = blkdiag(zeros(ny_pi),zeros(nu*N)); % No noise

[H, Lwc_pi, Jopt_pi] = optmeas(Gu, Gd, Juu, Jud, Wd, Wn);

H1 = H(:,1:ny_pi);
H2 = H(:,ny_pi+1:end);
K = inv(H2)*H1;
K0_pi = K(1:nu,:); % Gives the relation u = -K0*y, where y are the meas candidates

genPpi = calcP(Gpi,probPI);
J0_pi = clnorm(genPpi, -K0_pi);
%eig(Api - Bpi*K0_pi*Cpi)
[Kpi, Jpi] = clopt(Gpi, probPI, K0_pi, 'lft');

%% P controller
Qp = Q;
Rp = R;
Pp = Qp; %Pp = dlyap(Ad',Qp);
Wp = eye(nx);
Vp = zeros(ny);
probP = probStruct(Qp,Rp,Pp,N,Wp,Vp);

% Optimal measurement combination for P control
[Juu, Jud] = calcJuu(Gdisc, probP);

Gu = [Dd zeros(ny,nu*(N-1)); eye(nu*N)];
Gd = [Cd; zeros(nu*N,nx)];
  
Wd = eye(nx);
Wn = blkdiag(zeros(ny),zeros(nu*N)); % No noise

[H, Lwc_p, Jopt_p] = optmeas(Gu, Gd, Juu, Jud, Wd, Wn);

H1 = H(:,1:ny);
H2 = H(:,ny+1:end);
K = inv(H2)*H1;
K0_p = K(1:nu,:); % Gives the relation u = K0*y, where y are the meas candidates

genPp = calcP(Gdisc,probP);
J0_p = clnorm(genPp, -K0_p);
%eig(Ad - Bd*K0_p*Cd)
[Kp, Jp] = clopt(Gdisc, probP, K0_p, 'lft');

%% Closed-loop norms and performance indicators
Jlqr
Jpid
J0_pid
Jopt_pid
Lwc_pid
Jpd
J0_pd
Jopt_pd
Lwc_pd
Jpi
J0_pi
Jopt_pi
Lwc_pi
Jp
J0_p
Jopt_p
Lwc_p

% probPID2 = probPID;
% probPID2.Q = blkdiag(Q, diag([0 0]));
% probPID2.P = probPID2.Q;
% genP_pid = calcP(Gpid,probPID2);
% Jpid2 = clnorm(genP_pid, -Kpid);
% 
% probPD2 = probPD;
% probPD2.Q = blkdiag(Q, 0);
% probPD2.P = probPD2.Q;
% genP_pd = calcP(Gpd,probPD2);
% Jpd2 = clnorm(genP_pd, -Kpd);
% 
% probPI2 = probPI;
% probPI2.Q = blkdiag(Q, 0);
% probPI2.P = probPI2.Q;
% genP_pi = calcP(Gpi,probPI2);
% Jpi2 = clnorm(genP_pi, -Kpi);
% 
% Jp2 = Jp;

% Jlqr
% Jpid2
% Jpd2
% Jpi2
% Jp2

%% Simulation using the optimal controller parameters
time = 0:0.05:20;
N = length(time);

x0 = [1 0]';
Xlqr = zeros(nx,N+1); Xlqr(:,1) = x0;
Xpid = zeros(nx_pid,N+1); Xpid(:,1) = [x0; 0; 0];
Xpd = zeros(nx_pd,N+1); Xpd(:,1) = [x0; 0];
Xpi = zeros(nx_pi,N+1); Xpi(:,1) = [x0; 0];
Xp = zeros(nx,N+1); Xp(:,1) = x0;

Ylqr = zeros(ny,N);
Ypid = zeros(ny_pid,N);
Ypd = zeros(ny_pd,N);
Ypi = zeros(ny_pi,N);
Yp = zeros(ny,N);

Ulqr = zeros(nu,N);
Upid = zeros(nu,N);
Upd = zeros(nu,N);
Upi = zeros(nu,N);
Up = zeros(nu,N);

for i = 1:N
    % LQR controller
    Ylqr(i) = Cd*Xlqr(:,i);
    Ulqr(i) = -Klqr*Xlqr(:,i);
    Xlqr(:,i+1) = Ad*Xlqr(:,i) + Bd*Ulqr(i);
    
    % PID controller
    Ypid(:,i) = Cpid*Xpid(:,i);
    Upid(i) = -Kpid*Ypid(:,i);
    Xpid(:,i+1) = Apid*Xpid(:,i) + Bpid*Upid(i);
    
    % PD controller
    Ypd(:,i) = Cpd*Xpd(:,i);
    Upd(i) = -Kpd*Ypd(:,i);
    Xpd(:,i+1) = Apd*Xpd(:,i) + Bpd*Upd(i);
    
    % PI controller
    Ypi(:,i) = Cpi*Xpi(:,i);
    Upi(i) = -Kpi*Ypi(:,i);
    Xpi(:,i+1) = Api*Xpi(:,i) + Bpi*Upi(i);

    % P controller
    Yp(:,i) = Cd*Xp(:,i);
    Up(i) = -Kp*Yp(:,i);
    Xp(:,i+1) = Ad*Xp(:,i) + Bd*Up(i);
end

%% Plots the simulation results
close all

figure()
plot(time,Ylqr, '-b', 'Linewidth', 1.5)
hold on
plot(time,Ypid(1,:), '-k', 'Linewidth', 1.5)
plot(time,Ypd(1,:), '--k', 'Linewidth', 1.5)
plot(time,Ypi(1,:), '-.k', 'Linewidth', 1.5)
plot(time,Yp(1,:), ':k', 'Linewidth', 1.5)
hold off
xlim([0 10]);
ylim([-0.5 0.81]);
ylabel('Output');
xlabel('Time [s]');
l = legend('$K_{LQR}$', '$K_{PID}$', '$K_{PD}$', '$K_{PI}$', '$K_{P}$');
set(l, 'Interpreter', 'Latex')
set(l, 'FontSize', 12)
