% This example presents the closed-loop norms of a P, PI, PID, and LQR
% controller on a second-order system. It is expected that the LQR and PID
% controller has the same norm, while the PI controller has a higher norm,
% followed by the P controller.

% Note: adding integrators makes control design difficult as the
% closed-loop norm can become infinite

clc
clear all
close all
warning off all

Kp = -0.1:0.04:2;
Ki = -0.1:0.04:2;
Kd = -0.1:0.04:2;

Jvec_pid = zeros(length(Kp), length(Ki), length(Kd));
Jvec_pd = zeros(length(Kp), length(Kd));
Jvec_pi = zeros(length(Kp), length(Ki));
Jvec_p = zeros(length(Kp),1);

% The general second-order system is described as 
% g(s) = k/(tau^2 s^2 + 2 tau zeta s + 1)
% where tau = 1 and zeta = [0,1]
% zeta < 1 gives an underdamped system
% k = 1;
tau = 1;
zeta = 0.2;
g = tf(1, [tau^2, 2*tau*zeta, 1]);
G = ss(g); eig(G)
A = G.A; B = G.B;
C = G.C; D = G.D;
[nx,nu,ny] = sysdim(A,B,C);

% Discretization
% Note that d2c with zoh does not handle poles at zero z = 0, or negative
% real poles
Ts = 0.1;
Gdisc = c2d(G,Ts); eig(Gdisc)
Ad = Gdisc.A; Bd = Gdisc.B;
Cd = Gdisc.C; Dd = Gdisc.D;

% Problem formulation
Q = Cd'*Cd; %Q = diag([1 1]);
%Q = eye(2);
R = 1;
N = 100; % Horizon length
P = Q;
%P = dlyap(Ad',Q); % Solves the Lyapunov equation P = A'PA + Q

%% PID controller with limited derivative action
epsilon = 0.1; % Filter coefficient, numerical problems when set too low!

Gpid = sysaug(G, 'PID', 0, 'pos', epsilon);
Gpid = c2d(Gpid, Ts); % Discretizes

Apid = Gpid.A; Bpid = Gpid.B;
Cpid = Gpid.C; Dpid = Gpid.D;

[nx_pid, nu_pid, ny_pid] = sysdim(Apid, Bpid, Cpid);

% Note: if the integral action is not weighted, the marginal poles are not
% moved and the closed-loop system norm becomes Inf!
Qpid = Cpid'*diag([1 1e-3 0])*Cpid;
Rpid = R;
Ppid = Qpid; %Ppid = dlyap(Apid',Qpid);
Wpid = blkdiag(eye(nx),zeros(ny+ny)); % only weight on process states
Vpid = zeros(ny_pid);

probPID = probStruct(Qpid,Rpid,Ppid,N,Wpid,Vpid);

% Optimal measurement combination for PID control
[Juu, Jud] = calcJuu(Gpid, probPID);

Gu = [Dpid zeros(ny_pid,nu*(N-1));
      eye(nu*N)
      ];
Gd = [Cpid;
      zeros(nu*N,nx_pid)
      ];

Wd = eye(nx_pid);
Wn = blkdiag(zeros(ny_pid),zeros(nu*N)); % No noise

[H, Lwc_pid, Jopt_pid] = optmeas(Gu, Gd, Juu, Jud, Wd, Wn);

H1 = H(:,1:ny_pid);
H2 = H(:,ny_pid+1:end);
K = inv(H2)*H1;
K0_pid = K(1:nu,:); % Gives the relation u = K0*y, where y are the meas candidates

genPpid = calcP(Gpid,probPID);
%[Kpid, Jpid] = clopt(Gpid, probPID, K0_pid, 'lft');

for i1 = 1:length(Kp)
    for i2 = 1:length(Ki)
        for i3 = 1:length(Kd)
            Jvec_pid(i1,i2,i3) = clnorm(genPpid, -[Kp(i1) Ki(i2) Kp(i3)]);
        end
    end
end

%% PD controller with limited derivative action
epsilon = 0.1; % Filter coefficient, numerical problems when set too low!

Gpd = sysaug(G, 'PD', 0, 'pos', epsilon);
Gpd = c2d(Gpd, Ts); % Discretizes

Apd = Gpd.A; Bpd = Gpd.B;
Cpd = Gpd.C; Dpd = Gpd.D;

[nx_pd, nu_pd, ny_pd] = sysdim(Apd, Bpd, Cpd);

% Note: if the integral action is not weighted, the marginal poles are not
% moved and the closed-loop system norm becomes Inf!
Qpd = Cpd'*diag([1 0])*Cpd;
Rpd = R;
Ppd = Qpd; %Ppid = dlyap(Apid',Qpid);
Wpd = blkdiag(eye(nx),zeros(ny)); % only weight on process states
Vpd = zeros(ny_pd);

probPD = probStruct(Qpd,Rpd,Ppd,N,Wpd,Vpd);

% Optimal measurement combination for PID control
[Juu, Jud] = calcJuu(Gpd, probPD);

Gu = [Dpd zeros(ny_pd,nu*(N-1));
      eye(nu*N)
      ];
Gd = [Cpd;
      zeros(nu*N,nx_pd)
      ];

Wd = eye(nx_pd);
Wn = blkdiag(zeros(ny_pd),zeros(nu*N)); % No noise

[H, Lwc_pd, Jopt_pd] = optmeas(Gu, Gd, Juu, Jud, Wd, Wn);

H1 = H(:,1:ny_pd);
H2 = H(:,ny_pd+1:end);
K = inv(H2)*H1;
K0_pd = K(1:nu,:); % Gives the relation u = K0*y, where y are the meas candidates

genPpd = calcP(Gpd,probPD);

for i1 = 1:length(Kp)
    for i2 = 1:length(Kd)
        Jvec_pd(i1,i2) = clnorm(genPpd, -[Kp(i1) Kd(i2)]);
    end
end

%% PI controller
% Api = [A zeros(nx,ny);
%        C zeros(ny,ny)
%        ];
% Bpi = [B; D];
% Cpi = blkdiag(C, eye(ny));
% Dpi = [D; zeros(ny,nu)];
% 
% Gpi = ss(Api,Bpi,Cpi,Dpi); eig(Gpi)
Gpi = sysaug(G, 'PI', 0, 'pos');
Gpi = c2d(Gpi, Ts); % Discretizes

Api = Gpi.A; Bpi = Gpi.B;
Cpi = Gpi.C; Dpi = Gpi.D;

[nx_pi, nu_pi, ny_pi] = sysdim(Api, Bpi, Cpi);

% Note: if the integral action is not weighted, the marginal poles are not
% moved and the closed-loop system norm becomes Inf!
Qpi = blkdiag(Q,1e-4);
Qpi = Cpi'*diag([1 1e-3])*Cpi;
Rpi = R;
Ppi = Qpi; %Ppi = dlyap(Api',Qpi);
Wpi = blkdiag(eye(nx), zeros(ny));
Vpi = zeros(ny_pi);
probPI = probStruct(Qpi,Rpi,Ppi,N,Wpi,Vpi);

% Optimal measurement combination for PI control
[Juu, Jud] = calcJuu(Gpi, probPI);

Gu = [Dpi zeros(ny_pi,nu*(N-1));
      eye(nu*N)
      ];
Gd = [Cpi;
      zeros(nu*N,nx_pi)
      ];

Wd = eye(nx_pi);
Wn = blkdiag(zeros(ny_pi),zeros(nu*N)); % No noise

[H, Lwc_pi, Jopt_pi] = optmeas(Gu, Gd, Juu, Jud, Wd, Wn);

H1 = H(:,1:ny_pi);
H2 = H(:,ny_pi+1:end);
K = inv(H2)*H1;
K0_pi = K(1:nu,:); % Gives the relation u = -K0*y, where y are the meas candidates

genPpi = calcP(Gpi,probPI);

for i1 = 1:length(Kp)
    for i2 = 1:length(Ki)
        Jvec_pi(i1,i2) = clnorm(genPpi, -[Kp(i1) Ki(i2)]);
    end
end

%% P controller
Qp = Q;
Rp = R;
Pp = Qp; %Pp = dlyap(Ad',Qp);
Wp = eye(nx);
Vp = zeros(ny);
probP = probStruct(Qp,Rp,Pp,N,Wp,Vp);

% Optimal measurement combination for P control
[Juu, Jud] = calcJuu(Gdisc, probP);

Gu = [Dd zeros(ny,nu*(N-1)); eye(nu*N)];
Gd = [Cd; zeros(nu*N,nx)];

Wd = eye(nx);
Wn = blkdiag(zeros(ny),zeros(nu*N)); % No noise

[H, Lwc_p, Jopt_p] = optmeas(Gu, Gd, Juu, Jud, Wd, Wn);

H1 = H(:,1:ny);
H2 = H(:,ny+1:end);
K = inv(H2)*H1;
K0_p = K(1:nu,:); % Gives the relation u = K0*y, where y are the meas candidates

genPp = calcP(Gdisc,probP);

for i1 = 1:length(Kp)
	Jvec_p(i1) = clnorm(genPp, -Kp(i1));
end


%% Calculate global minimas
