% This example presents the closed-loop norms of a P, PI, PID, and LQR
% controller on a second-order system. It is expected that the LQR and PID
% controller has the same norm, while the PI controller has a higher norm,
% followed by the P controller.

% Note: adding integrators makes control design difficult as the
% closed-loop norm can become infinite

clc
clear all
close all

Zeta = 0.1:0.1:2;
Kpid_vec = zeros(length(Zeta), 3);
Jpid_vec = zeros(length(Zeta), 1);
J0_pid_vec = zeros(length(Zeta), 1);
Lwc_pid_vec = zeros(length(Zeta), 1);
margin_vec = zeros(length(Zeta), 4);

for i = 1:length(Zeta)

% The general second-order system is described as 
% g(s) = k/(tau^2 s^2 + 2 tau zeta s + 1)
% where tau = 1 and zeta = [0,1]
% zeta < 1 gives an underdamped system
k = 1;
tau = 1;
zeta = Zeta(i);
g = tf(k, [tau^2, 2*tau*zeta, 1]);
G = ss(g); eig(G)
A = G.A; B = G.B;
C = G.C; D = G.D;
[nx,nu,ny] = sysdim(A,B,C);

% Discretization
% Note that d2c with zoh does not handle poles at zero z = 0, or negative
% real poles
Ts = 0.1;
Gdisc = c2d(G,Ts); eig(Gdisc)
Ad = Gdisc.A; Bd = Gdisc.B;
Cd = Gdisc.C; Dd = Gdisc.D;

% Problem formulation
Q = Cd'*Cd; %Q = diag([1 1]);
%Q = eye(2);
R = 1;
N = 100; % Horizon length
P = Q;
%P = dlyap(Ad',Q); % Solves the Lyapunov equation P = A'PA + Q

%% PID controller with limited derivative action
epsilon = 0.1; % Filter coefficient, numerical problems when set too low!

% Apid = [A zeros(nx,ny+ny);
%         C zeros(ny,ny+ny);
%         (1/epsilon)*C*A zeros(ny) -(1/epsilon)*eye(ny)
%         ];
% Bpid = [B; D; (epsilon*C*B-D)/(epsilon^2)];
% Cpid = blkdiag(C, eye(ny+ny)); 
% Dpid = [D; zeros(ny,nu); D/epsilon];

%Gpid = ss(Apid,Bpid,Cpid,Dpid); eig(Gpid) 

Gpid = sysaug(G, 'PID', 0, 'pos', epsilon);
Gpid = c2d(Gpid, Ts); % Discretizes

Apid = Gpid.A; Bpid = Gpid.B;
Cpid = Gpid.C; Dpid = Gpid.D;

[nx_pid, nu_pid, ny_pid] = sysdim(Apid, Bpid, Cpid);

% Note: if the integral action is not weighted, the marginal poles are not
% moved and the closed-loop system norm becomes Inf!
Qpid = Cpid'*diag([1 1e-3 0])*Cpid; % 1e-3
Rpid = R;
Ppid = Qpid; %Ppid = dlyap(Apid',Qpid);
Wpid = blkdiag(eye(nx),zeros(ny+ny)); % only weight on process states
Vpid = zeros(ny_pid);

probPID = probStruct(Qpid,Rpid,Ppid,N,Wpid,Vpid);

% Optimal measurement combination for PID control
[Juu, Jud] = calcJuu(Gpid, probPID);

Gu = [Dpid zeros(ny_pid,nu*(N-1));
      eye(nu*N)
      ];
Gd = [Cpid;
      zeros(nu*N,nx_pid)
      ];
  
Wd = eye(nx_pid);
Wn = blkdiag(zeros(ny_pid),zeros(nu*N)); % No noise

[H, Lwc_pid] = optmeas(Gu, Gd, Juu, Jud, Wd, Wn);

H1 = H(:,1:ny_pid);
H2 = H(:,ny_pid+1:end);
K = inv(H2)*H1;
K0_pid = K(1:nu,:); % Gives the relation u = K0*y, where y are the meas candidates

genPpid = calcP(Gpid,probPID);
J0_pid = clnorm(genPpid, -K0_pid);
%eig(Apid - Bpid*K0_pid*Cpid)
[Kpid, Jpid] = clopt(Gpid, probPID, K0_pid, 'lft');

%% Stores data
Kpid_vec(i,:) = Kpid;
Jpid_vec(i,1) = Jpid;
J0_pid_vec(i,1) = J0_pid;
Lwc_pid_vec(i,1) = Lwc_pid;

[Gm,Pm,Wg,Wp] = margin(Kpid*Gpid);
margin_vec(i,:) = [Gm,Pm,Wg,Wp];

end

%% Plots controller param as a function of zeta
figure
plot(Zeta, Kpid_vec(:,1), 'k', 'Linewidth', 2);
hold on
plot(Zeta, Kpid_vec(:,2), 'b', 'Linewidth', 2);
plot(Zeta, Kpid_vec(:,3), 'g', 'Linewidth', 2);
hold off
xlim([Zeta(1) Zeta(end)])
xlabel('\zeta');
ylabel('PID controller parameters');
l = legend('$K_p$', '$K_i$', '$K_d$');
set(l, 'Interpreter', 'Latex')
set(l, 'FontSize', 12)

figure
plot(Zeta, margin_vec(:,1), 'k', 'Linewidth', 2)
hold on
plot(Zeta, margin_vec(:,2), 'b', 'Linewidth', 2)
plot(Zeta, margin_vec(:,3), 'g', 'Linewidth', 2)
hold off
xlim([Zeta(1) Zeta(end)])
xlabel('\zeta');
ylabel('Open-loop transfer function margins and bandwidth');
legend('Gain margin [abs]', 'Phase margin [deg]', 'Bandwidth [rad/sec]')

