% This example presents the closed-loop norms of a P, PI, PID, and LQR
% controller on a second-order system. It is expected that the LQR and PID
% controller has the same norm, while the PI controller has a higher norm,
% followed by the P controller.

% Note: adding integrators makes control design difficult as the
% closed-loop norm can become infinite

clc
clear all
close all
warning off all

Zeta = 0.08; %[0.2 0.5 0.8];
Kp = -1:0.01:5;
Ki = -0.2:0.01:2;

K0_1 = [0 0];
K0_2 = [0 0];
K0_3 = [0 0];
J0_1 = 0;
J0_2 = 0;
J0_3 = 0;
J_vec1 = zeros(length(Kp), length(Ki));
J_vec2 = zeros(length(Kp), length(Ki));
J_vec3 = zeros(length(Kp), length(Ki));

for k = 1:length(Zeta) %max length is three
    
    % The general second-order system is described as 
    % g(s) = k/(tau^2 s^2 + 2 tau zeta s + 1)
    % where tau = 1 and zeta = [0,1]
    % zeta < 1 gives an underdamped system
    % k = 1;
    tau = 1;
    zeta = Zeta(k);
    g = tf(1, [tau^2, 2*tau*zeta, 1]);
    G = ss(g); eig(G)
    A = G.A; B = G.B;
    C = G.C; D = G.D;
    [nx,nu,ny] = sysdim(A,B,C);

    % Discretization
    % Note that d2c with zoh does not handle poles at zero z = 0, or negative
    % real poles
    Ts = 0.1;
    Gdisc = c2d(G,Ts); eig(Gdisc)
    Ad = Gdisc.A; Bd = Gdisc.B;
    Cd = Gdisc.C; Dd = Gdisc.D;

    % Problem formulation
    Q = Cd'*Cd; %Q = diag([1 1]);
    %Q = eye(2);
    R = 1;
    N = 100; % Horizon length
    P = Q;
    %P = dlyap(Ad',Q); % Solves the Lyapunov equation P = A'PA + Q

    % PI controller
    Gpi = sysaug(G, 'PI', 0, 'pos');
    Gpi = c2d(Gpi, Ts); % Discretizes

    Api = Gpi.A; Bpi = Gpi.B;
    Cpi = Gpi.C; Dpi = Gpi.D;

    [nx_pi, nu_pi, ny_pi] = sysdim(Api, Bpi, Cpi);

    % Note: if the integral action is not weighted, the marginal poles are not
    % moved and the closed-loop system norm becomes Inf!
    Qpi = Cpi'*diag([1 1e-3])*Cpi;
    Rpi = R;
    Ppi = Qpi; %Ppi = dlyap(Api',Qpi);
    Wpi = blkdiag(eye(nx), zeros(ny));
    Vpi = zeros(ny_pi);
    probPI = probStruct(Qpi,Rpi,Ppi,N,Wpi,Vpi);

    % Optimal measurement combination for PI control
    [Juu, Jud] = calcJuu(Gpi, probPI);

    Gu = [Dpi zeros(ny_pi,nu*(N-1));
          eye(nu*N)
          ];
    Gd = [Cpi;
          zeros(nu*N,nx_pi)
          ];

    Wd = eye(nx_pi);
    Wn = blkdiag(zeros(ny_pi),zeros(nu*N)); % No noise

    [H, Lwc_pi, Jopt_pi] = optmeas(Gu, Gd, Juu, Jud, Wd, Wn);

    H1 = H(:,1:ny_pi);
    H2 = H(:,ny_pi+1:end);
    K = inv(H2)*H1;
    K0 = K(1:nu,:); % Gives the relation u = -K0*y, where y are the meas candidates
    K0
    
    genPpi = calcP(Gpi,probPI);
    J0 = clnorm(genPpi, -K0);
 
    if (k == 1)
        K0_1 = K0;
        J0_1 = J0;
    elseif (k == 2)
        K0_2 = K0;
        J0_2 = J0;   
    else
        K0_3 = K0;
        J0_3 = J0;    
    end
    
    % calculates closed-loop objective function value for different
    % controller parameters
    for i = 1:length(Kp)
        for j = 1:length(Ki)
            
            %closed-loop optimization
            Jnow = clnorm(genPpi, -[Kp(i) Ki(j)]);
            if(k == 1)
                J_vec1(i,j) = Jnow;
            elseif (k == 2)
                J_vec2(i,j) = Jnow;
            else
                J_vec3(i,j) = Jnow;
            end
            
        end
    end
end

%% Plots controller param as a function of zeta
close all

figure
mesh(Ki, Kp, J_vec1)
hold on
contour(Ki, Kp, J_vec1)
plot3(K0_1(2), K0_1(1), J0_1, 'Marker', 'o', 'MarkerSize', 12, 'MarkerEdgeColor', 'k', 'MarkerFaceColor', 'w', 'Linewidth', 1.5)
plot3(K0_1(2), K0_1(1), J0_1, 'Marker', 'x', 'MarkerSize', 12, 'MarkerEdgeColor', 'k', 'Linewidth', 1.5)
hold off

xlim([Ki(1) Ki(end)])
ylim([Kp(1) Kp(end)])
zlim([0 10])

lx = xlabel('$K_i$');
set(lx, 'Interpreter', 'Latex')
set(lx, 'FontSize', 12)
ly = ylabel('$K_p$');
set(ly, 'Interpreter', 'Latex')
set(ly, 'FontSize', 12)
lz = zlabel('$||F_l(P,K)||_2$');
set(lz, 'Interpreter', 'Latex')
set(lz, 'FontSize', 12)
%legend('K_p', 'K_i', 'K_d');

figure
mesh(Ki, Kp, J_vec2)
hold on
contour(Ki, Kp, J_vec2)
plot3(K0_2(2), K0_2(1), J0_2, 'Marker', 'o', 'MarkerSize', 12, 'MarkerEdgeColor', 'k', 'MarkerFaceColor', 'w', 'Linewidth', 1.5)
plot3(K0_2(2), K0_2(1), J0_2, 'Marker', 'x', 'MarkerSize', 12, 'MarkerEdgeColor', 'k', 'Linewidth', 1.5)
hold off

xlim([Ki(1) Ki(end)])
ylim([Kp(1) Kp(end)])
zlim([0 10])

lx = xlabel('$K_i$');
set(lx, 'Interpreter', 'Latex')
set(lx, 'FontSize', 12)
ly = ylabel('$K_p$');
set(ly, 'Interpreter', 'Latex')
set(ly, 'FontSize', 12)
lz = zlabel('$||F_l(P,K)||_2$');
set(lz, 'Interpreter', 'Latex')
set(lz, 'FontSize', 12)
%legend('K_p', 'K_i', 'K_d');

figure
mesh(Ki, Kp, J_vec3)
hold on 
contour(Ki, Kp, J_vec3)
plot3(K0_3(2), K0_3(1), J0_3, 'Marker', 'o', 'MarkerSize', 12, 'MarkerEdgeColor', 'k', 'MarkerFaceColor', 'w', 'Linewidth', 1.5)
plot3(K0_3(2), K0_3(1), J0_3, 'Marker', 'x', 'MarkerSize', 12, 'MarkerEdgeColor', 'k', 'Linewidth', 1.5)
hold off

xlim([Ki(1) Ki(end)])
ylim([Kp(1) Kp(end)])
zlim([0 10])

lx = xlabel('$K_i$');
set(lx, 'Interpreter', 'Latex')
set(lx, 'FontSize', 12)
ly = ylabel('$K_p$');
set(ly, 'Interpreter', 'Latex')
set(ly, 'FontSize', 12)
lz = zlabel('$||F_l(P,K)||_2$');
set(lz, 'Interpreter', 'Latex')
set(lz, 'FontSize', 12)
%legend('K_p', 'K_i', 'K_d');