% Controller design for the udaq
% Low order controllers (P, PI, PID) are measured up against the LQG with
% integral effect

clc
%clear all
close all

udaq_model; % loads model
udaq_csys = d2c(udaq_sys); % Continuous model
[Ac,Bc,Cc,Dc] = ssdata(udaq_csys);
[nx,nu,ny] = sysdim(A,B,C); % dimensions

% Design weights
Q = diag([25 1e-3 15 5 1e-3]); % States have been scaled
Qi = diag([.01 .01]);     % Integrator weights
Qd = diag([1e-4 1e-4]);   % Derivative weights (Remember 1/epsilon from sysaug)
R = diag([.1 2 1]);      % Input weights
P = Q;
W = eye(nx);            % Process noise weights (on states) 
V = eye(ny);          % Measurement noise weights
%V = diag([1e-4 1e-4]);

N = 120; % Horizon length

%% LQG design

% 1) LQR design for augmented system (sigma = int(y))
Aint = [Ac zeros(nx,ny); Cc zeros(ny)]; %approximates integrator states 
Bint = [Bc; Dc];
Cint = [Cc zeros(ny,ny); zeros(ny,nx) eye(ny)];
Dint = [Dc; zeros(ny,nu)];
Gint = ss(Aint, Bint, Cint, Dint);
Gint = c2d(Gint, Ts);
[Aint, Bint, Cint, Dint] = ssdata(Gint);

%Qlqr = Cint'*diag([Qy1, Qy2, 1e-1, 1e-1])*Cint; % 1,1,1e-1,1e-1
Qlqr = blkdiag(Q, Qi); % State + integrator weights
Rlqr = R;
[Kr, Plqr] = dlqr(Aint, Bint, Qlqr, Rlqr);
Krx = Kr(:,1:nx); Kri = Kr(:,nx+1:end);

% 2) Steady-state Kalman filter design
% Model on the form:
% x(k+1) = A x(k) + B u(k) + Bnoise w(k)
% y(k) = C x(k) + Dnoise w(k) + v(k)
Bnoise = eye(nx);
Dnoise = zeros(ny,nx);
est_mdl = ss(A, [B Bnoise], C, [D Dnoise], Ts); % Estimator model
[Kest, Kf, Pf, Mf] = kalman(est_mdl, W, V);

% 3) Form overall controller, u = Klqg*y 
Alqg = [A-A*Mf*C-B*Krx -B*Kri; zeros(ny,nx) eye(ny)];
Blqg = [A*Mf; Ts*eye(ny)];
Clqg = -[Krx Kri];
Dlqg = zeros(nu,ny);
Klqg = ss(Alqg, Blqg, Clqg, Dlqg, Ts);

% Calculates closed-loop norms
probLQG = probStruct(Q,R,P,N,W,V); 
Plqg = calcP(udaq_sys, probLQG); % Generalized plant with noise weights
Jlqg = clnorm(Plqg, Klqg)
eig(Plqg.A)

%% PID design

% Augments system with PID controller states
epsilon = 2; % Filter coefficient
Gpid = sysaug(udaq_csys, 'PID', 0, 'pos', epsilon); 

% Discretizes the augmented system
Gpid = c2d(Gpid, Ts); 
[Apid,Bpid,Cpid,Dpid] = ssdata(Gpid);
[nx_pid, nu_pid, ny_pid] = sysdim(Apid, Bpid, Cpid);

% PID design weights
Qpid = blkdiag(Q, Qi, Qd);
Rpid = R;
Ppid = Qpid; %Ppid = dlyap(Apid',Qpid); Not used
Wpid = blkdiag(W, eye(ny+ny)); % only weight on process states
Vpid = blkdiag(V, zeros(ny), V); % Same noise weight on P, I, and D outputs
probPID = probStruct(Qpid,Rpid,Ppid,N,Wpid,Vpid);

% Optimal measurement combination for PID control
[Juu, Jud] = calcJuu(Gpid, probPID);

Gu = [Dpid zeros(ny_pid,nu*(N-1));
      eye(nu*N)
      ];
Gd = [Cpid;
      zeros(nu*N,nx_pid)
      ];

Wd = Wpid;
Wn = blkdiag(Vpid, zeros(nu*N)); % No noise

[H, Lwc_pid, Jopt_pid] = optmeas(Gu, Gd, Juu, Jud, Wd, Wn);

% Extracts first invariant
% Gives the relation u = K0*y, where y are the meas candidates
H1 = H(:,1:ny_pid);
H2 = H(:,ny_pid+1:end);
K = inv(H2)*H1;
K0_pid = K(1:nu,:); 

genPpid = calcP(Gpid,probPID);
J0_pid = clnorm(genPpid, -K0_pid);
eig(Apid - Bpid*K0_pid*Cpid)
K0_pid
J0_pid

[Kpid, Jpid] = clopt(Gpid, probPID, K0_pid, 'lft');
Kp = Kpid(:,1:ny); Ki = Kpid(:,ny+1:ny+ny); Kd = Kpid(:,ny+ny+1:end);
Jpid

% %% PI design
% % Augments system with PI controller states
% Gpi = sysaug(udaq_csys, 'PI', 0, 'pos'); 
% 
% % Discretizes the augmented system
% Gpi = c2d(Gpi, Ts); 
% [Api,Bpi,Cpi,Dpi] = ssdata(Gpi);
% [nx_pi, nu_pi, ny_pi] = sysdim(Api, Bpi, Cpi);
% 
% % PID design weights
% Qpi = blkdiag(Q, Qi);
% Rpi = R;
% Ppi = Qpi; %Ppid = dlyap(Apid',Qpid); Not used
% Wpi = blkdiag(W, eye(ny)); % only weight on process states
% Vpi = blkdiag(V, zeros(ny)); % Same noise weight on P, I, and D outputs
% probPI = probStruct(Qpi,Rpi,Ppi,N,Wpi,Vpi);
% 
% % Optimal measurement combination for PID control
% [Juu, Jud] = calcJuu(Gpi, probPI);
% 
% Gu = [Dpi zeros(ny_pi,nu*(N-1));
%       eye(nu*N)
%       ];
% Gd = [Cpi;
%       zeros(nu*N,nx_pi)
%       ];
% 
% Wd = eye(nx_pi); % should maybe use W?
% Wn = blkdiag(Vpi, zeros(nu*N)); % No noise
% 
% [H, Lwc_pi, Jopt_pi] = optmeas(Gu, Gd, Juu, Jud, Wd, Wn);
% 
% H1 = H(:,1:ny_pi);
% H2 = H(:,ny_pi+1:end);
% K = inv(H2)*H1;
% K0_pi = K(1:nu,:); % Gives the relation u = K0*y, where y are the meas candidates
% 
% genPpi = calcP(Gpi,probPI);
% J0_pi = clnorm(genPpi, -K0_pi);
% eig(Api - Bpi*K0_pi*Cpi)
% K0_pi
% J0_pi
% 
% [Kpi, Jpi] = clopt(Gpi, probPI, K0_pi, 'lft');
% Kpi_p = Kpi(:,1:ny); Kpi_i = Kpi(:,ny+1:end);
% Jpi

%% Filtered PI design
% Augments system with PI controller states
%Gpi = sysaug(udaq_csys, 'PI', 0, 'pos');
tau = 2; % Filter coeff
Api2 = [Ac zeros(nx, 2*ny); 
        Cc -(1/tau)*eye(ny) zeros(ny); 
        zeros(ny,nx) eye(ny) zeros(ny)];
Bpi2 = [Bc; Dc; zeros(ny,nu)];
Cpi2 = [zeros(ny,nx) eye(ny) zeros(ny); zeros(ny,nx+ny) eye(ny)];
Dpi2 = zeros(2*ny, nu);
Gpi2 = ss(Api2, Bpi2, Cpi2, Dpi2);

% Discretizes the augmented system
Gpi2 = c2d(Gpi2, Ts); 
[Api2,Bpi2,Cpi2,Dpi2] = ssdata(Gpi2);
[nx_pi2, nu_pi2, ny_pi2] = sysdim(Api2, Bpi2, Cpi2);

% PI design weights
Qpi2 = blkdiag(Q, 1e-3*Qi, Qi);
Rpi2 = R;
Ppi2 = Qpi2; %Ppid = dlyap(Apid',Qpid); Not used
Wpi2 = blkdiag(W, eye(2*ny)); % only weight on process states
Vpi2 = blkdiag(V, zeros(ny)); % Same noise weight on P, I, and D outputs
probPI2 = probStruct(Qpi2,Rpi2,Ppi2,N,Wpi2,Vpi2);

% Optimal measurement combination for PID control
[Juu, Jud] = calcJuu(Gpi2, probPI2);

Gu = [Dpi2 zeros(ny_pi2,nu*(N-1));
      eye(nu*N)
      ];
Gd = [Cpi2;
      zeros(nu*N,nx_pi2)
      ];

Wd = Wpi2;
Wn = blkdiag(Vpi2, zeros(nu*N));

[H, Lwc_pi2, Jopt_pi2] = optmeas(Gu, Gd, Juu, Jud, Wd, Wn);

% Extracts first invariant
% Gives the relation u = K0*y, where y are the meas candidates
H1 = H(:,1:ny_pi2);
H2 = H(:,ny_pi2+1:end);
K = inv(H2)*H1;
K0_pi2 = K(1:nu,:); 

genPpi2 = calcP(Gpi2,probPI2);
J0_pi2 = clnorm(genPpi2, -K0_pi2);
eig(Api2 - Bpi2*K0_pi2*Cpi2)
K0_pi2
J0_pi2

[Kpi2, Jpi2] = clopt(Gpi2, probPI2, K0_pi2, 'lft');
Kpi2_p = Kpi2(:,1:ny); Kpi2_i = Kpi2(:,ny+1:end);
Jpi2

%% Optimization done
udaq_controller_run = 1;