% Controller design for the udaq
% Low order controllers (P, PI, PID) are measured up against the LQG with
% integral effect

clc
clear all
close all

udaq_model; % loads model

%Modifies model, u3 = alpha*u2
alpha = 1;
U = [1 0; 0 1; 0 alpha];
B = B*U;
D = D*U;
udaq_sys = ss(A,B,C,D,Ts);

udaq_csys = d2c(udaq_sys); % Continuous model
[Ac,Bc,Cc,Dc] = ssdata(udaq_csys);
[nx,nu,ny] = sysdim(A,B,C); % dimensions

% Design weights
Q = diag([25 1e-3 15 2 1e-3]);  % States have been scaled
Qi = diag([.005 .01]);          % Integrator weights
Qd = diag([1e-4 1e-4]);         % Derivative weights (Remember 1/epsilon from sysaug)
R = diag([.1 2]);               % Input weights
P = Q;
W = eye(nx);                    % Process noise weights (on states) 
V = eye(ny);                    % Measurement noise weights

N = 120; % Horizon length

%% LQG design

% 1) LQR design for augmented system (sigma = int(y))
Aint = [Ac zeros(nx,ny); Cc zeros(ny)]; %approximates integrator states 
Bint = [Bc; Dc];
Cint = [Cc zeros(ny,ny); zeros(ny,nx) eye(ny)];
Dint = [Dc; zeros(ny,nu)];
Gint = ss(Aint, Bint, Cint, Dint);
Gint = c2d(Gint, Ts);
[Aint, Bint, Cint, Dint] = ssdata(Gint);

Qlqr = blkdiag(Q, Qi); % State + integrator weights
Rlqr = R;
[Kr, Plqr] = dlqr(Aint, Bint, Qlqr, Rlqr);
Krx = Kr(:,1:nx); Kri = Kr(:,nx+1:end);

% 2) Steady-state Kalman filter design
Bnoise = eye(nx);
Dnoise = zeros(ny,nx);
est_mdl = ss(A, [B Bnoise], C, [D Dnoise], Ts); % Estimator model
[Kest, Kf, Pf, Mf] = kalman(est_mdl, W, V);

% 3) Form overall controller, u = Klqg*y 
Alqg = [A-A*Mf*C-B*Krx -B*Kri; zeros(ny,nx) eye(ny)];
Blqg = [A*Mf; Ts*eye(ny)];
Clqg = -[Krx Kri];
Dlqg = zeros(nu,ny);
Klqg = ss(Alqg, Blqg, Clqg, Dlqg, Ts);

% Calculates closed-loop norms
probLQG = probStruct(Q,R,P,N,W,V); 
Plqg = calcP(udaq_sys, probLQG); % Generalized plant with noise weights
Jlqg = clnorm(Plqg, Klqg)
eig(Plqg.A)

%LQG norm: 77.0215
%PID norm: 83.7597

%% PID design

% Augments system with PID controller states
epsilon = 2; % Filter coefficient
Gpid = sysaug(udaq_csys, 'PID', 0, 'pos', epsilon); 

% Discretizes the augmented system
Gpid = c2d(Gpid, Ts); 
[Apid,Bpid,Cpid,Dpid] = ssdata(Gpid);
[nx_pid, nu_pid, ny_pid] = sysdim(Apid, Bpid, Cpid);

% PID design weights
Qpid = blkdiag(Q, Qi, Qd);
Rpid = R;
Ppid = Qpid;
Wpid = blkdiag(W, eye(ny+ny)); 
Vpid = blkdiag(V, zeros(ny), V); % Same noise weight on P, I, and D outputs
probPID = probStruct(Qpid,Rpid,Ppid,N,Wpid,Vpid);

% Optimal measurement combination for PID control
[Juu, Jud] = calcJuu(Gpid, probPID);

Gu = [Dpid zeros(ny_pid,nu*(N-1));
      eye(nu*N)
      ];
Gd = [Cpid;
      zeros(nu*N,nx_pid)
      ];

Wd = Wpid;
Wn = blkdiag(Vpid, zeros(nu*N)); % No noise

[H, Lwc_pid, Jopt_pid] = optmeas(Gu, Gd, Juu, Jud, Wd, Wn);

% Extracts first invariant
% Gives the relation u = K0*y, where y are the meas candidates
H1 = H(:,1:ny_pid);
H2 = H(:,ny_pid+1:end);
K = inv(H2)*H1;
K0_pid = K(1:nu,:); 

genPpid = calcP(Gpid,probPID);
J0_pid = clnorm(genPpid, -K0_pid);
eig(Apid - Bpid*K0_pid*Cpid)
K0_pid
J0_pid

[Kpid, Jpid] = clopt(Gpid, probPID, K0_pid, 'lft');
Kp = Kpid(:,1:ny); Ki = Kpid(:,ny+1:ny+ny); Kd = Kpid(:,ny+ny+1:end);
Jpid

%% Filtered PI design
% Augments system with PI controller states
%Gpi = sysaug(udaq_csys, 'PI', 0, 'pos');
tau = 2; % Filter coeff
Api2 = [Ac zeros(nx, 2*ny); 
        Cc -(1/tau)*eye(ny) zeros(ny); 
        zeros(ny,nx) eye(ny) zeros(ny)];
Bpi2 = [Bc; Dc; zeros(ny,nu)];
Cpi2 = [zeros(ny,nx) eye(ny) zeros(ny); zeros(ny,nx+ny) eye(ny)];
Dpi2 = zeros(2*ny, nu);
Gpi2 = ss(Api2, Bpi2, Cpi2, Dpi2);

% Discretizes the augmented system
Gpi2 = c2d(Gpi2, Ts); 
[Api2,Bpi2,Cpi2,Dpi2] = ssdata(Gpi2);
[nx_pi2, nu_pi2, ny_pi2] = sysdim(Api2, Bpi2, Cpi2);

% PI design weights
Qpi2 = blkdiag(Q, 1e-3*Qi, Qi);
Rpi2 = R;
Ppi2 = Qpi2; %Ppid = dlyap(Apid',Qpid); Not used
Wpi2 = blkdiag(W, eye(2*ny)); % only weight on process states
Vpi2 = blkdiag(V, zeros(ny)); % Same noise weight on P, I, and D outputs
probPI2 = probStruct(Qpi2,Rpi2,Ppi2,N,Wpi2,Vpi2);

% Optimal measurement combination for PID control
[Juu, Jud] = calcJuu(Gpi2, probPI2);

Gu = [Dpi2 zeros(ny_pi2,nu*(N-1));
      eye(nu*N)
      ];
Gd = [Cpi2;
      zeros(nu*N,nx_pi2)
      ];

Wd = Wpi2;
Wn = blkdiag(Vpi2, zeros(nu*N));

[H, Lwc_pi2, Jopt_pi2] = optmeas(Gu, Gd, Juu, Jud, Wd, Wn);

% Extracts first invariant
% Gives the relation u = K0*y, where y are the meas candidates
H1 = H(:,1:ny_pi2);
H2 = H(:,ny_pi2+1:end);
K = inv(H2)*H1;
K0_pi2 = K(1:nu,:); 

genPpi2 = calcP(Gpi2,probPI2);
J0_pi2 = clnorm(genPpi2, -K0_pi2);
eig(Api2 - Bpi2*K0_pi2*Cpi2)
K0_pi2
J0_pi2

[Kpi2, Jpi2] = clopt(Gpi2, probPI2, K0_pi2, 'lft');
Kpi2_p = Kpi2(:,1:ny); Kpi2_i = Kpi2(:,ny+1:end);
Jpi2

%% Optimization done
udaq_controller_2inputs_run = 1;

%% Plots singular values for the closed-loop systems
close all

N = 200;
omega_max = pi/Ts;
omega = logspace(-4,log10(omega_max),N);

scale = diag([1.6e-3 1.5e-4]);

% LQG
Llqg = udaq_sys*Klqg; % Loop tf
Slqg = inv(eye(2) + Llqg); % Sensitivity tf
sigmas_lqg = sigma(Slqg,omega);
sigmas_lqg = 20*log10(sigmas_lqg); % dB

% PID
Lpid = Kpid*Gpid;
Spid = inv(eye(2) + Lpid);
sigmas_pid = sigma(Spid,omega);
sigmas_pid = 20*log10(sigmas_pid); % dB

% PI
Lpi = Kpi2*Gpi2;
Spi = inv(eye(2) + Lpi);
sigmas_pi = sigma(Spi,omega);
sigmas_pi = 20*log10(sigmas_pi); % dB


figure();
semilogx(omega,sigmas_lqg(1,:),'k-', 'Linewidth', 1.5);
hold on;
semilogx(omega,sigmas_lqg(2,:),'k--', 'Linewidth', 1.5);
semilogx(omega,zeros(1,length(omega)), 'r-');
xlim([omega(1) omega(end)]);
ylim([-100 20]);
legend('$\overline{\sigma}(S_{LQG})$', '$\underline{\sigma}(S_{LQG})$', '$0$ [dB]');
ylabel('Magnitude [dB]')
xlabel('Frequency [rad/s]');

figure();
semilogx(omega,sigmas_pid(1,:),'k-', 'Linewidth', 1.5);
hold on;
semilogx(omega,sigmas_pid(2,:),'k--', 'Linewidth', 1.5);
semilogx(omega,zeros(1,length(omega)), 'r-');
xlim([omega(1) omega(end)]);
ylim([-100 20]);
legend('$\overline{\sigma}(S_{PID})$', '$\underline{\sigma}(S_{PID})$', '$0$ [dB]');
ylabel('Magnitude [dB]')
xlabel('Frequency [rad/s]');

figure();
semilogx(omega,sigmas_pi(1,:),'k-', 'Linewidth', 1.5);
hold on;
semilogx(omega,sigmas_pi(2,:),'k--', 'Linewidth', 1.5);
semilogx(omega,zeros(1,length(omega)), 'r-');
xlim([omega(1) omega(end)]);
ylim([-100 20]);
legend('$\overline{\sigma}(S_{PI})$', '$\underline{\sigma}(S_{PI})$', '$0$ [dB]');
ylabel('Magnitude [dB]')
xlabel('Frequency [rad/s]');


figure();
semilogx(omega,scale*sigmas_lqg(1,:),'k-', 'Linewidth', 1.5);
hold on;
semilogx(omega,sigmas_pid(1,:),'k--', 'Linewidth', 1.5);
semilogx(omega,sigmas_pi(1,:),'k-.', 'Linewidth', 1.5);
semilogx(omega,zeros(1,length(omega)), 'r-');
xlim([omega(1) omega(end)]);
ylim([-100 20]);
legend('$\bar{\sigma}(S_{LQR})$', '$\bar{\sigma}(S_{PID})$', '$\bar{\sigma}(S_{PI})$', '$0$ [dB]');
ylabel('Magnitude [dB]')
xlabel('Frequency [rad/s]');

% figure();
% bode(Slqg)