function [ H, Lwc, Jopt ] = optmeas( Gu, Gd, Juu, Jud, Wd, Wn )
%OPTMEAS Exact local method for finding the optimal measurement combination
%   c = H*y, where H is a linear mapping
%   H = argmin [maximum singular value of M(H)] 
%   H = argmin norm(H*Ftilde) subject to H*Gy = Juu^(1/2)
%   
%   Ouputs:
%   H       - Optimal linear measurement combination
%   Lwc     - Worst case loss associated with H
%   Jopt    - Optimal objective function value 
%             Jopt = norm(Hopt*Ftilde,'fro');
%
%   Inputs:
%   Gu,Gd   - Measurement candidates: y = Gu*u + Gd*d
%   Juu,Jud - Sensitivity matrices
%   Wd,Wn   - Disturbance and noise scaling matrices

display('Finding optimal measurement combination H...');

% dimensions
ny = size(Gu, 1);   % Number of measurements
nu = size(Gu, 2);   % Number of inputs
nd = size(Gd, 2);   % Number of disturbances

% Finds F and Ftilde
F = -(Gu*inv(Juu)*Jud - Gd);
Ftilde = [F*Wd Wn];

% Checks which method to use based on number of measurements
% If nc = nu we control as many variables as we have inputs

if(fullrank(Ftilde*Ftilde')) % If Ftilde*Ftilde' is non-singular (i.e. has full rank)
    
    % Explicit expression for H
    H = inv(Ftilde*Ftilde') * Gu * inv(Gu'*inv(Ftilde*Ftilde')*Gu) * sqrtm(Juu);
    H = H';
    
else % General case, Ftilde*Ftilde' singular. Uses CVX
    
    % Convex optimization, assumes that nc = nu
    cvx_begin
        variable H(nu, ny);
        minimize norm(H*Ftilde, 'fro');
        subject to
            H*Gu == sqrtm(Juu);
    cvx_end
    
    % Optimization gives same answer
    % J = inline('norm(Ht''*Ftilde)','Ht','Ftilde'); 
    % 
    % H0 = eye(size(Juu,1), ny); % Initial guess on H
    % 
    % [Ht, Lwc] = fmincon(@(Ht) J(Ht,Ftilde), H0', [], [], Gu', sqrtm(Juu)');
    % H2 = Ht';
    % H2
    
end

% Optimal objective function value
Jopt = norm(H*Ftilde, 'fro');

% Calculates worst case loss
Md = -sqrtm(Juu)*inv(H*Gu)*H*F*Wd;
Mn = -sqrtm(Juu)*inv(H*Gu)*H*Wn;
M = [Md Mn];
Lwc = 0.5*svds(M,1)^2;

end
