function [ Gaug ] = sysaug( varargin )
%SYSAUG Augments a continuous system with controller (PI, PD, or PID) dynamics
%   A continuous model G is augmented with the dynamics of a PI, PD, or PID
%   controller: u(s) = Kp*e + Ki*(1/s)*e + Kp*(s/(epsilon*s + 1))*e
%   With the error defined as: e = r - y, r denoting the reference signal. 
%   The PID controller (Kpid) is assumed to be on parallel
%   form and u = Kpid*e. 
%
%   The 'feedback' parameter can be set to negative ('neg', default): 
%   u = Kpid*e, or positive ('pos'): u = -Kpid*e. Also,
%   the PID controller is implemented with limited derivative action by 
%   including a low-pass filter with a time-constant 'epsilon' 
%   (default value is epsilon = 0.1). In the MIMO case the same filter is
%   used on all SISO controllers. The reference signal is omitted by
%   setting 'ref' to 0 (default value is 1), which yields e = -y.
%
%   Examples:
%   [ Gaug ] = sysaug( G ): Augments G with a the default PID controller
%   [ Gaug ] = sysaug( G, 'PI' ): Augments G with a PI controller
%   [ Gaug ] = sysaug( G, 'PI', ref, 'feedback' ): Augments G with positive
%   feedback PI controller
%   [ Gaug ] = sysaug( G, 'PD', ref, 'feedback' ): Augments G with positive 
%   feedback PD controller 
%   PI controller. ref decides if to reference signal is included.
%   [ Gaug ] = sysaug( G, 'PID' ): Augments G with a PID controller
%   [ Gaug ] = sysaug( G, 'PID', ref, 'feedback', epsilon): Augments G with a PID
%   controller, with the derivative filter specified by epsilon and negative/positive feedback.

G = varargin{1}; % Continuous model
method = 'PID'; % PID is the default controller
ref = 1; % Reference signal included
feedback = 'neg'; % Negative feedback is set as default
epsilon = 0.1; % Default value for low-pass filter

if (nargin >= 2)
    method = varargin{2};
    if (~strcmp(method, 'PI') && ~strcmp(method, 'PID') && ~strcmp(method, 'PD'))
        error('Unknown controller. Please choose PI, PD, or PID.');
    end
end

if (nargin >= 3)
    ref = varargin{3};
end

if (nargin >= 4)
    feedback = varargin{4};
    if (~strcmp(feedback, 'neg') && ~strcmp(feedback, 'pos'))
        error('Unknown feedback specification. Please choose neg or pos.');
    end
end

if (nargin == 5)
	epsilon = varargin{5};
end

% Augments the system
A = G.A; B = G.B; C = G.C; D = G.D;
[nx,nu,ny] = sysdim(A,B,C);

% PID controller
if (strcmp(method, 'PID'))
    
    Aaug = [A zeros(nx,ny+ny);
            -C zeros(ny,ny+ny);
            C*A zeros(ny) -(1/epsilon)*eye(ny)
            ];
        
    if (ref == 0)
        Baug = [B; -D; (1/epsilon)*(epsilon*C*B - D)];
        Daug = [-D; zeros(ny,nu); -(1/epsilon)*D];
    else
        Baug = [B zeros(nx,ny); 
                -D eye(ny); 
                (1/epsilon)*(epsilon*C*B - D) (1/epsilon)*eye(ny)
                ];
        Daug = [-D eye(ny); 
                zeros(ny,nu+ny);
                -(1/epsilon)*D (1/epsilon)*eye(ny)
                ];
    end
    
    Caug = [-C zeros(ny,ny+ny);
            zeros(ny,nx) eye(ny) zeros(ny);
            zeros(ny,nx+ny) -(1/epsilon)*eye(ny)
            ];
        
    if (strcmp(feedback, 'pos'))
        Caug = -Caug;
        Daug = -Daug;
    end
    
end

% PI controller
if (strcmp(method, 'PI'))
    
    Aaug = [A zeros(nx,ny);
            -C zeros(ny,ny);
            ];
      
    if (ref == 0)
        Baug = [B; -D];
        Daug = [-D; zeros(ny,nu)];
    else
        Baug = [B zeros(nx,ny); 
                -D eye(ny)
                ];
        Daug = [-D eye(ny); 
                zeros(ny,nu+ny)
                ];
    end
    
    Caug = [-C zeros(ny,ny);
            zeros(ny,nx) eye(ny)
            ];
        
    if (strcmp(feedback, 'pos'))
        Caug = -Caug;
        Daug = -Daug;
    end
    
end

% PD controller
if (strcmp(method, 'PD'))
    
    Aaug = [A zeros(nx,ny);
            C*A -(1/epsilon)*eye(ny)
            ];
        
    if (ref == 0)
        Baug = [B; (1/epsilon)*(epsilon*C*B - D)];
        Daug = [-D; -(1/epsilon)*D];
    else
        Baug = [B zeros(nx,ny); 
                (1/epsilon)*(epsilon*C*B - D) (1/epsilon)*eye(ny)
                ];
        Daug = [-D eye(ny);
                -(1/epsilon)*D (1/epsilon)*eye(ny)
                ];
    end
    
    Caug = [-C zeros(ny,ny);
            zeros(ny,nx) -(1/epsilon)*eye(ny)
            ];
        
    if (strcmp(feedback, 'pos'))
        Caug = -Caug;
        Daug = -Daug;
    end
    
end

Gaug = ss(Aaug, Baug, Caug, Daug);
